from transformers import AutoModelForCausalLM, AutoTokenizer

from rustic_ai.core.agents.commons.message_formats import (
    GenerationPromptRequest,
    GenerationPromptResponse,
)
from rustic_ai.core.guild import agent
from rustic_ai.core.guild.agent import Agent, AgentMode, AgentSpec, AgentType
from rustic_ai.huggingface.agents.models import PyTorchAgentProps


class PhiAgentProps(PyTorchAgentProps):
    model_id: str = "microsoft/phi-2"


class LLMPhiAgent(Agent[PhiAgentProps]):
    """An Agent that generates a response to the given generation prompt using Phi 2."""

    def __init__(
        self,
        agent_spec: AgentSpec[PhiAgentProps],
    ) -> None:
        super().__init__(
            agent_spec=agent_spec,
            agent_type=AgentType.BOT,
            agent_mode=AgentMode.LOCAL,
        )
        if agent_spec.properties is None:
            agent_spec.properties = PhiAgentProps()

        self._device = agent_spec.properties.torch_device
        model_id = agent_spec.properties.model_id
        self.model = AutoModelForCausalLM.from_pretrained(model_id, trust_remote_code=True).to(self._device).eval()
        self.tokenizer = AutoTokenizer.from_pretrained(model_id, trust_remote_code=True, torch_dtype="auto")

    @agent.processor(GenerationPromptRequest)
    def generate_prompt(self, ctx: agent.ProcessContext[GenerationPromptRequest]) -> None:
        """
        Generates a response to the given generation prompt.
        """
        generation_prompt = ctx.payload.generation_prompt
        inputs = self.tokenizer(generation_prompt, return_tensors="pt", return_attention_mask=False).to(self._device)

        outputs = self.model.generate(**inputs, max_length=200)
        text = self.tokenizer.batch_decode(outputs)[0]

        ctx.send(GenerationPromptResponse(generation_prompt=generation_prompt, generated_response=text))



from pydantic import BaseModel

from rustic_ai.core.agents.commons.message_formats import ErrorMessage
from rustic_ai.core.guild import agent
from rustic_ai.core.guild.agent import Agent, AgentMode, AgentSpec, AgentType
from rustic_ai.huggingface.inference_mixin import (
    HuggingfaceInferenceMixin,
)


class QuestionWithContext(BaseModel):
    context: str
    question: str


class SquadAnswer(BaseModel):
    text: str


class SquadAgent(Agent, HuggingfaceInferenceMixin):
    def __init__(
        self,
        agent_spec: AgentSpec,
    ) -> None:
        super().__init__(
            agent_spec,
            AgentType.BOT,
            AgentMode.REMOTE,
        )

    @agent.processor(QuestionWithContext)
    async def find_answer(self, ctx: agent.ProcessContext[QuestionWithContext]):
        """
        Uses the Huggingface model (deepset/roberta-base-squad2) to find the answer to the question in the context.
        """
        question_with_context = ctx.payload
        payload = self.generate_input_prompt(question_with_context.model_dump())
        response = await self.run_inference("deepset/roberta-base-squad2", payload)

        if response:
            result = SquadAnswer(text=response["answer"])

            ctx.send(result)

        else:
            ctx.send(
                ErrorMessage(
                    agent_type=self.get_qualified_class_name(),
                    error_type="InferenceError",
                    error_message="No answer received from the model.",
                )
            )  # pragma: no cover



from typing import List

from pydantic import BaseModel, Field

from rustic_ai.core.agents.commons import GenerationPromptRequest
from rustic_ai.core.agents.commons.media import MediaLink
from rustic_ai.core.guild import BaseAgentProps


class PyTorchAgentProps(BaseAgentProps):
    """
    Properties for the HuggingFace PyTorch agent.
    """

    torch_device: str = Field(default="cpu", title="Torch device to use for inference")
    model_id: str = Field(title="HuggingFace Model ID to use")


class ImageGenerationRequest(GenerationPromptRequest):
    """
    Represents a request for generating images using a given generation model.

    Args:
        num_images: The number of images to generate. Default is 1.
        height: The height of the generated images. Default is 512.
        width: The width of the generated images. Default is 512.
        num_inference_steps: The number of inference steps for generating the images. Default is 50.
        image_format: The format of the generated images. Default is "png".
        guidance_scale: The scale used for guiding the image generation. Default is 7.5.
    """

    num_images: int = Field(default=1)
    height: int = Field(default=512)
    width: int = Field(default=512)
    num_inference_steps: int = Field(default=50)
    image_format: str = Field(default="png")
    guidance_scale: float = Field(default=7.5)


class ImageGenerationResponse(BaseModel):
    files: List[MediaLink] = Field(default_factory=list)
    errors: List[str] = Field(default_factory=list)
    request: str = Field(default_factory=str)



import mimetypes
import uuid

import soundfile as sf
import torch
from datasets import load_dataset
from transformers import pipeline

from rustic_ai.core.agents.commons.media import MediaLink
from rustic_ai.core.agents.commons.message_formats import (
    ErrorMessage,
    GenerationPromptRequest,
)
from rustic_ai.core.guild import agent
from rustic_ai.core.guild.agent import Agent, AgentMode, AgentType
from rustic_ai.core.guild.agent_ext.depends.filesystem import FileSystem
from rustic_ai.core.guild.dsl import AgentSpec


class SpeechT5TTSAgent(Agent):

    def __init__(
        self,
        agent_spec: AgentSpec,
    ) -> None:
        super().__init__(
            agent_spec,
            AgentType.BOT,
            AgentMode.LOCAL,
        )
        self._synthesiser = pipeline("text-to-speech", "microsoft/speecht5_tts")

        embeddings_dataset = load_dataset("Matthijs/cmu-arctic-xvectors", split="validation")
        self._speaker_embedding = torch.tensor(embeddings_dataset[7306]["xvector"]).unsqueeze(0)

    @agent.processor(GenerationPromptRequest, depends_on=["filesystem:guild_fs:True"])
    def convert(self, ctx: agent.ProcessContext[GenerationPromptRequest], guild_fs: FileSystem) -> None:
        """
        Handles the received message.

        Args:
            message (Message): The received message.
        """
        data = ctx.payload
        try:
            speech = self._synthesiser(
                data.generation_prompt, forward_params={"speaker_embeddings": self._speaker_embedding}
            )
            filename = f"{uuid.uuid4()}.wav"
            try:
                with guild_fs.open(filename, "wb") as f:
                    sf.write(f, speech["audio"], samplerate=speech["sampling_rate"])
                    output = MediaLink(
                        url=filename,
                        name=filename,
                        metadata={"sampling_rate": speech["sampling_rate"]},
                        on_filesystem=True,
                        mimetype=mimetypes.guess_type(filename)[0],
                    )
                    ctx.send(output)
            except Exception as e:
                ctx.send(
                    ErrorMessage(
                        agent_type=self.get_qualified_class_name(), error_type="FileWriteError", error_message=str(e)
                    )
                )
        except Exception as ex:
            ctx.send(
                ErrorMessage(
                    agent_type=self.get_qualified_class_name(),
                    error_type="SpeechGenerationError",
                    error_message=str(ex),
                )
            )



import mimetypes
import uuid

import torch
from diffusers import (
    EulerAncestralDiscreteScheduler,
    StableDiffusionInstructPix2PixPipeline,
)
from PIL import Image, ImageOps

from rustic_ai.core.agents.commons.media import MediaLink
from rustic_ai.core.agents.commons.message_formats import ErrorMessage
from rustic_ai.core.guild import agent
from rustic_ai.core.guild.agent import Agent, ProcessContext
from rustic_ai.core.guild.agent_ext.depends.filesystem import FileSystem
from rustic_ai.core.guild.dsl import AgentSpec
from rustic_ai.huggingface.agents.models import (
    ImageGenerationRequest,
    ImageGenerationResponse,
    PyTorchAgentProps,
)


class ImageQuery(ImageGenerationRequest):
    image_path: str


class Pix2PixProps(PyTorchAgentProps):
    model_id: str = "timbrooks/instruct-pix2pix"


class Image2ImageAgent(Agent[Pix2PixProps]):
    """
    Generate image from a given image and a query
    """

    def __init__(self, agent_spec: AgentSpec[Pix2PixProps]):
        super().__init__(agent_spec)
        if agent_spec.properties is None:
            agent_spec.properties = Pix2PixProps()
        self.torch_device = agent_spec.properties.torch_device
        model_id = agent_spec.properties.model_id

        self._pipeline = StableDiffusionInstructPix2PixPipeline.from_pretrained(model_id, safety_checker=None)
        self._pipeline.scheduler = EulerAncestralDiscreteScheduler.from_config(self._pipeline.scheduler.config)
        self._pipeline.to(torch.device(self.torch_device))

    @agent.processor(ImageQuery, depends_on=["filesystem:guild_fs:True"])
    def process_user_query(self, ctx: ProcessContext[ImageQuery], guild_fs: FileSystem) -> None:
        """
        Process the users query
        """
        image_gen_request = ctx.payload

        try:
            with guild_fs.open(image_gen_request.image_path, "rb") as f:
                loaded_image = Image.open(f)
                image = ImageOps.exif_transpose(loaded_image)
                if image is not None:
                    image = image.convert("RGB")
                else:
                    raise ValueError("Image cannot be none")

            prompt = [image_gen_request.generation_prompt] * image_gen_request.num_images
            image_input = [image] * image_gen_request.num_images
            output_images = self._pipeline(
                prompt,
                image=image_input,
                num_inference_steps=image_gen_request.num_inference_steps,
                guidance_scale=image_gen_request.guidance_scale,
                height=image_gen_request.height,
                width=image_gen_request.width,
            ).images

            result = ImageGenerationResponse(files=[], errors=[], request=image_gen_request.model_dump_json())
            for output_image in output_images:
                filename = f"{uuid.uuid4()}.{image_gen_request.image_format}"
                try:
                    with guild_fs.open(filename, "wb") as f:
                        output_image.save(f)
                    media_link = MediaLink(
                        url=filename, name=filename, mimetype=mimetypes.guess_type(filename)[0], on_filesystem=True
                    )
                    result.files.append(media_link)
                except Exception as e:
                    result.errors.append(f"Failed to write image file {filename}:{e}")
            ctx.send(result)
        except Exception as ex:
            ctx.send(
                ErrorMessage(
                    agent_type=self.get_qualified_class_name(),
                    error_type="ImageGenerationError",
                    error_message=str(ex),
                )
            )



import json
import logging
import os
from urllib.parse import urljoin

from rustic_ai.core.agents.utils.http_client import HttpClientMixin


class HuggingfaceInferenceMixin(HttpClientMixin):

    async def run_inference(
        self, model: str, prompt_payload: str, inference_endpoint: str = "https://api-inference.huggingface.co/models/"
    ):
        api_url = urljoin(inference_endpoint, model)

        headers = {}
        if not model or not isinstance(model, str):
            logging.info("model not provided")  # pragma: no cover
            return
        if not prompt_payload or not isinstance(prompt_payload, str):
            logging.info("prompt not provided")  # pragma: no cover
            return

        huggingface_api_key = os.getenv("HUGGINGFACE_API_KEY", "")
        if huggingface_api_key:
            headers["Authorization"] = f"Bearer {huggingface_api_key}"

        try:
            response = await self.send_json(api_url, extra_headers=headers, json_data=prompt_payload)
            return response
        except Exception as e:
            logging.error(f"Error running inference: {e}")
            return None

    def generate_input_prompt(self, data: dict) -> str:
        return json.dumps({"inputs": data, "options": {"wait_for_model": True}})
